/*
  ggHiggs code based on

    R. D. Ball, M. Bonvini, S. Forte, S. Marzani, G. Ridolfi
    arXiv:1303.3590

  please cite it if you use it
*/

#include <iomanip>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdio.h>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/stat.h> 
#include <sys/time.h> 
#include <getopt.h> 

#include "./src/ggHiggs.hh"



using namespace std;
using namespace ggHiggs;


/*
  If you like, you can customize your alpha_s routine.
  However, consistency with PDF sets suggests the use of the LHAPDF alphas routine.
*/
double ggHiggs::alphas(double mu) {
  //return alphasPDF(mu);
  return alphas_ihixs(mu, 3, alphasPDF(GetZmass()), GetZmass()); // taken from ihixs code
  //return alphas_ihixs(mu, order, LHAPDF::alphasPDF(mZ), mZ); // taken from ihixs code
}





void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -p  --pdf-set <set>    specify PDFs being used (default \"NNPDF23_nnlo_as_0118\")" << endl
       << " -m  --mass <mass>      set Higgs mass [GeV]" << endl
       << " -E  --Ecm <energy>     set collider energy [GeV]" << endl
       << " -f  --fact <murat>     set muF/mu  ratio" << endl
       << " -r  --ren <murat>      set muR/muF ratio" << endl
       << " -t  --topmass <mass>   set top quark mass [GeV] (either mt_pole or mt(mt), dependeing on the -S flag)" << endl
       << " -b  --include-bottom   include the bottom Yukawa coupling" << endl
       << " -c  --include-charm    include the charm  Yukawa coupling" << endl
       << " -S  --MSbarMass        use MSbar running top mass (with mt(mt) = 162.7 GeV)" << endl
       << " -e  --noEFT            use finite quark masses (default: infinite-mtop EFT approximation)" << endl
       << endl;
  cout << "Best setup:" << endl
       << "  ggHiggs-verbose -e -b -c" << endl
       << endl;
  exit(0);
  return;
}






double muFrat = 1;
double muRrat = 1;
const string sorder[4] = {"LO", "NLO", "NNLO", "NNNLO"};
void read_arguments(int argc, char* argv[], double &mH, double &sqrts, string &set, bool &nob, bool &noc) {
  const char* const short_options = "ht:ep:f:r:m:E:bcS";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "topmass", 1, NULL, 't' },
					 { "noEFT", 0, NULL, 'e' },
					 { "pdf-set", 1, NULL, 'p' },
					 { "fact", 1, NULL, 'f' },
					 { "ren", 1, NULL, 'r' },
					 { "mass", 1, NULL, 'm' },
					 { "Ecm", 1, NULL, 'E' },
					 { "include-bottom", 0, NULL, 'b' },
					 { "include-charm", 0, NULL, 'c' },
					 { "MSbarMass", 0, NULL, 'S' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'e':
      finite_mt = true;
      break;
    case 't':
      mtop = strtod(optarg, NULL);
      mtmt = mtop;
      break;
    case 'p':
      sset << optarg;
      set = sset.str();
      break;
    case 'f':
      _muF_mu_ratio = strtod(optarg, NULL);
      break;
    case 'r':
      _muR_muF_ratio = strtod(optarg, NULL);
      break;
    case 'm':
      mH = strtod(optarg, NULL);
      break;
    case 'E':
      sqrts = strtod(optarg, NULL);
      break;
    case 'b':
      nob = false;
      break;
    case 'c':
      noc = false;
      break;
    case 'S':
      MSbarMass = true;
      cout << "\033[0;31m WARNING: MSbar scheme for the top mass is a new feature, implemented only for the top in the large-mt EFT at the moment. Use with care.\033[0m" << endl;
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  //
  return;
}







int main (int argc, char* argv[]) {

  quiet = false;

  // relative precision for integrals
  INTEGRAL_PRECISION = 1.e-3;



  // default values (modified by runtime arguments)
  double mH = 125;
  double sqrts = 13000;
  string set = "PDF4LHC15_nnlo_mc";


  // read runtime arguments
  bool nob=true, noc=true; // when true, bottom and charm do NOT run into the loop coupling to the Higgs
  read_arguments(argc, argv, mH, sqrts, set, nob, noc);
  _muF_mu_ratio  = muFrat;
  _muR_muF_ratio = muRrat/muFrat;


  // Start time counter
  struct timeval t0, t1;
  gettimeofday(&t0,NULL);



  // init PDFs
  cout << "Loading PDF set " << set << endl;
  int subset = 0;
  pdf_init(set, subset);

  // masses (negative mass means the quark will not contribute to the loop)
  mbot = (nob ? -1 : 1) * getQuarkMass(5);
  mcha = (noc ? -1 : 1) * getQuarkMass(4);


  // init
  double tau = mH*mH/sqrts/sqrts;
  //sxc = new SmallxCoeffs("./src/MBDFV.dat"); // old - deprecated (valid only for a single particle running in the loop)
  sxc = new SmallxCoeffs(mH,mtop,mbot,mcha);
  init_consts(mH);
  SetScaleDependence(1,1);
  //
  cout << "Higgs mass:       " << mH << " GeV" << endl
       << "Collider energy:  " << sqrts << " GeV" << endl
       << "tau = mH^2/s =    " << tau << endl
       << "top    mass =     " << mtop << "  running in the loop" << endl
       << "bottom mass =     " << fabs(mbot) << "  " << (nob ? "NOT " : "") << "running in the loop" << endl
       << "charm  mass =     " << fabs(mcha) << "  " << (noc ? "NOT " : "") << "running in the loop" << endl
       << "quark mass effects beyond LO: " << (finite_mt ? "YES (exact theory)" : "NO (EFT)") << endl
       << "Renormalization scale muR = " << _muR_muF_ratio*_muF_mu_ratio << " * mH" << endl
       << "Factorization   scale muF = " << _muF_mu_ratio << " * mH" << endl
       << "alphas(muR) = " << alphas(_muR_muF_ratio*_muF_mu_ratio*mH) << endl;
  //





  /******* RUN EXAMPLE *******/

  // Note also that the same PDF set has to be used at every order.
  double LO=1, NLO=0, NLOgg=0, NNLO=0, NNLOgg=0, NNNLO=0, NNNLOgg=0;
  //
  double LOpref = xs_LO_prefactor(mH);
  double muF = mH *_muF_mu_ratio;
  double muR = muF*_muR_muF_ratio;
  double alpha_s = alphas(muR);
  // results
  double rexact[3], rexact_gg[3], rsoft2[3], rsoft1[3], rzsoft[3], rNsoft[3], rhighenergy[3], rhighenergyLL[3], momconsunc[3];

  // LO
  LO = tau*Lum(tau,muF)*LOpref*pow(alpha_s,2);
  if(!quiet) cout << "sigma0 = " << LO << endl;

  // NLO
  order = 1;
  __all_channels__ = true;
  compute(mH, sqrts, LO, 0, 0, &rexact[0], NULL, NULL, NULL, NULL);
  NLO = LO + rexact[0];
  __all_channels__ = false;  // gg only
  compute(mH, sqrts, LO, 0, 0, &rexact_gg[0], &rsoft2[0], &rsoft1[0], &rNsoft[0], &rhighenergy[0], &momconsunc[0], &rhighenergyLL[0], &rzsoft[0]);
  //compute(mH, sqrts, LO, 0, 0, &rexact_gg[0], NULL, NULL, NULL, NULL);
  NLOgg = LO + rexact_gg[0];

  // NNLO
  order = 2;
  __all_channels__ = true;
  compute(mH, sqrts, LO, NLO, 0, &rexact[1], NULL, NULL, NULL, NULL);
  NNLO = NLO + rexact[1];
  __all_channels__ = false;
  compute(mH, sqrts, LO, NLOgg, 0, &rexact_gg[1], &rsoft2[1], &rsoft1[1], &rNsoft[1], &rhighenergy[1], &momconsunc[1], &rhighenergyLL[1], &rzsoft[1]);
  //compute(mH, sqrts, LO, NLOgg, 0, &rexact_gg[1], NULL, NULL, NULL, NULL);
  NNLOgg = NLOgg + rexact_gg[1];

  // NNNLO
  order = 3;
  if(!finite_mt) {
    __all_channels__ = true;
    //compute(mH, sqrts, LO, NLO, NNLO, &rexact[2], &rsoft2[2], &rsoft1[2], &rNsoft[2], &rhighenergy[2], &momconsunc[2], &rhighenergyLL[2], &rzsoft[2]);
    compute(mH, sqrts, LO, NLO, NNLO, &rexact[2], NULL, NULL, NULL, NULL, NULL, NULL, &rzsoft[2]);
  }
  NNNLO = NNLO + rexact[2];
  __all_channels__ = false;
  compute(mH, sqrts, LO, NLOgg, NNLOgg, &rexact_gg[2], &rsoft2[2], &rsoft1[2], &rNsoft[2], &rhighenergy[2], &momconsunc[2], &rhighenergyLL[2], &rzsoft[2]);
  NNNLOgg = NNLOgg + rexact_gg[2];





  // final results
  double appr1[3], appr2[3], appr_mean[3], ProcError1[3], ProcError2[3];
  double herr = 0.05; // momentum violation (fraction)
  for(int i=0; i<3; i++) {
    appr1[i] = (rsoft1[i]+rhighenergy[i]);
    appr2[i] = (rsoft2[i]+rhighenergy[i]);
    appr_mean[i] = appr2[i];
    ProcError1[i] = fabs(appr1[i]-appr2[i]);   // estimated error from soft uncertainty only
    ProcError2[i] = herr*fabs(momconsunc[i]);  // estimated error from high-energy uncertainties [momentum violation only]
    ProcError2[i] = sqrt(pow(herr*fabs(momconsunc[i]),2)+pow(rhighenergy[i]-rhighenergyLL[i],2));  // estimated error from high-energy uncertainties
  }
  //
  cout << endl
       << "The following cross-sections have been computed with the same PDF set (" << set << ") at every order."
       << endl << endl;
  cout << "LO:" << endl;
  cout << " exact      = " << LO << endl;
  cout << "NLO:" << endl;
  cout << " exact      = " << NLO << endl;
  cout << " exact(gg)  = " << NLOgg << endl;
  cout << " approx     = " << LO + appr_mean[0] << " +/- " << ProcError1[0] << "(soft) +/- " << ProcError2[0] << "(high-energy)" << endl;
  cout << " approx(gg) = " << LO + appr_mean[0] << " +/- " << ProcError1[0] << "(soft) +/- " << ProcError2[0] << "(high-energy)" << endl;
  cout << "NNLO:" << endl;
  cout << " exact      = " << NNLO << endl;
  cout << " exact(gg)  = " << NNLOgg << endl;
  cout << " approx     = " << NLO   + appr_mean[1] << " +/- " << ProcError1[1] << "(soft) +/- " << ProcError2[1] << "(high-energy)" << endl;
  cout << " approx(gg) = " << NLOgg + appr_mean[1] << " +/- " << ProcError1[1] << "(soft) +/- " << ProcError2[1] << "(high-energy)" << endl;
  cout << "NNNLO:" << endl;
  cout << " exact      = " << NNNLO << endl;
  cout << " exact(gg)  = " << NNNLOgg << endl;
  cout << " approx     = " << NNLO   + appr_mean[2] << " +/- " << ProcError1[2] << "(soft) +/- " << ProcError2[2] << "(high-energy)" << endl;
  cout << " approx(gg) = " << NNLOgg + appr_mean[2] << " +/- " << ProcError1[2] << "(soft) +/- " << ProcError2[2] << "(high-energy)" << endl;
  cout << endl;
  //

  double EW = 1+deltaEW(mH,mtop);
  cout << "ElectroWeak correction factor = " << EW << endl << endl;

  cout << "LO(exact) / LO(EFT) = " << LOpref/xs_LO_prefactor_EFT() << endl;

  // Stop time counter
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << endl << "Total time: " << t << "s" << endl;


  return 0;

}
